function Pim = classify_image(im,model,layer0,Pim)

im_lo = imfilter(im,fspecial('gaussian',5,1),'replicate');

[ly lx lz] = size(im);

if nargin < 3
    layer0 = 1;
end

for layer=layer0:model.numlayers
    if layer > 1
        Pim_lo  = imfilter(Pim,fspecial('gaussian',5,1),'replicate');
    end
    
    if strcmp(model.type,'lr')
        % HINT: use same trick used below for W instead of V
        if layer==1
            X = decompose_im_pca( im, model.buf_size,model.At{layer},model.mu{layer});
        else
            comps_im = (model.buf_size*2+1)^2*lz;
            X = decompose_im_pca(im,  model.buf_size, model.At{layer}(:,1:comps_im),     model.mu{layer}(1:comps_im)) + ...
                decompose_im_pca(Pim, model.buf_size, model.At{layer}(:,comps_im+1:end), model.mu{layer}(comps_im+1:end));
        end 
        P = exp(model.W{layer}*X);
        P = P./repmat(sum(P,1),size(P,1),1);
    elseif strcmp(model.type,'mlp')
        % cleverly premultiply the PCA matrix by the first layer of the
        % model for speed.  (assumes size(V,1) < size(At,1))
        V  = model.V{layer};
        W  = model.W{layer};
        nclasses = size(W,1);
        At = model.At{layer};
        mu = model.mu{layer};
        comps_im = (model.buf_size*2+1)^2*lz;
        comps_p  = (model.buf_size*2+1)^2*nclasses;
        cor1 =            1:comps_im;
        cor2 =   comps_im+1:2*comps_im;
        cor3 = 2*comps_im+1:2*comps_im+comps_p;
        cor4 = 2*comps_im+comps_p+1:length(mu);
        if layer==1
            %X = decompose_im_pca( im, model.buf_size,At,mu);
            %VX = decompose_im_pca( im, model.buf_size,V*At,mu);
            VX = decompose_im_pca(im,   model.buf_size  , V*At(:,cor1), mu(cor1),1) + ...
                 decompose_im_pca(im_lo,model.buf_size*2, V*At(:,cor2), mu(cor2),2);
        else
            %X = decompose_im_pca(im,  model.buf_size, At(:,1:comps_im),     mu(1:comps_im)) + ...
            %    decompose_im_pca(Pim, model.buf_size,
            %    At(:,comps_im+1:end), mu(comps_im+1:end));
            VX = decompose_im_pca(im,      model.buf_size  , V*At(:,cor1), mu(cor1),1) + ...
                 decompose_im_pca(im_lo,   model.buf_size*2, V*At(:,cor2), mu(cor2),2) + ...
                 decompose_im_pca(Pim,     model.buf_size  , V*At(:,cor3), mu(cor3),1) + ...
                 decompose_im_pca(Pim_lo,  model.buf_size*2, V*At(:,cor4), mu(cor4),2);
        end
        
        %sig = @(x) tanh(x);
        %P = exp(model.W{layer}*sig(model.V{layer}*X));
        %P = exp(W*tanh(VX));
        P = exp(W*mytanh_fast(VX)); % use my fast tabled tanh approximation
        P = P./repmat(sum(P,1),size(P,1),1);
    else
        error('unsupported model type');
    end
    
    Pim = zeros(ly,lx,size(P,1));
    for c=1:size(P,1)
        Pim(:,:,c) = reshape(P(c,:),ly,lx);
    end
    
    if nargout==0
        figure(1); imshow(im);
        figure(2); miximshow(Pim); title(layer); drawnow
    end
    
end

end