function Pim = classify_image_scaled(im,model,layer0,Pim)

num_scales    = model.num_scales;
num_pca_comps = model.num_pca_comps;

im_scaled = get_scalespace(im,num_scales);

[ly lx lz] = size(im);

if nargin < 3
    layer0 = 1;
end

% coordinates from different parts of image and prob space
comps_im = (model.buf_size*2+1)^2*lz;
comps_p  = (model.buf_size*2+1)^2*model.num_labels;
cor_im{1} = 1:comps_im;
for scale=2:num_scales
    last = max(cor_im{scale-1});
    cor_im{scale}=last+1:last+comps_im;
end
last = max(cor_im{end});
cor_p{1} = last+1:last+comps_p;
for scale=2:num_scales
    last = max(cor_p{scale-1});
    cor_p{scale}=last+1:last+comps_p;
end

for layer=layer0:model.numlayers
    At = model.At{layer};
    mu = model.mu{layer};
    
    if layer>1, P_scaled = get_scalespace(Pim,num_scales); end
    
    if strcmp(model.type,'lr')
        error('lr not currently supported');
    elseif strcmp(model.type,'dt')
        T  = model.T{layer};
        if layer==1, P_scaled=[]; end
        node = decompose_im_dt(im_scaled,P_scaled,model,layer);
        
        P = T.classprob(node,:)';
        P = .95*P + .05/model.num_labels; % avoid /0 errors
    elseif strcmp(model.type,'mlp')
        % cleverly premultiply the PCA matrix by the first layer of the
        % model for speed.  (assumes size(V,1) < size(At,1))
        V  = model.V{layer};
        myV = V(:,1:end-1); % no bias term
        bV  = V(:,end);     % bias term only
        W  = model.W{layer};
        
        VX = repmat(bV,1,ly*lx);
        
        if layer==1
            for scale=1:num_scales
                step = 2^(scale-1);
                VX = VX + decompose_im_pca(im_scaled{scale},model.buf_size*step, myV*At(:,cor_im{scale}), mu(cor_im{scale}),step);
            end
                 
        else
             for scale=1:num_scales
                step = 2^(scale-1);
                VX = VX + decompose_im_pca(im_scaled{scale},model.buf_size*step, myV*At(:,cor_im{scale}), mu(cor_im{scale}),step);
            end
            for scale=1:num_scales
                step = 2^(scale-1);
                 VX = VX + decompose_im_pca(P_scaled{scale} ,model.buf_size*step, myV*At(:,cor_p{scale}), mu(cor_p{scale}), step);
            end
        end
       
        P = exp(W*mytanh_fast(VX)); % use my fast tabled tanh approximation
        P = P./repmat(sum(P,1),size(P,1),1);
    else
        error('unsupported model type');
    end
    
    Pim = zeros(ly,lx,size(P,1));
    for c=1:size(P,1)
        Pim(:,:,c) = reshape(P(c,:),ly,lx);
    end
    
    if nargout==0
        figure(1); imshow(im);
        figure(1+layer); miximshow(Pim); title(layer); drawnow
    end
    
end

end