/* Simple test program for the RAD package. */

/* Compute f(n,x) = the product of x[0], x[1], ..., x[n-1] */
/* and the gradient of f for two x values, and print the result. */

//#include "rad.hpp"
#include <stdio.h>
#include "mex.h"   //--This one is required
#define nanoMatDebug false // 1 for safety/error checking; 0 for faster and unsafe(er)
#include "nanoMatT.cpp"
#include "math.h"
#define mmin(X,Y) ((X) < (Y) ? (X) : (Y))
#define mmax(X,Y) ((X) > (Y) ? (X) : (Y))

/*
typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<ADvar> adMat;
typedef ADvar var;
*/


typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<double> adMat;
typedef double var;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){
    int i, j, k, l, m, n, y;
    
    dMat  X  = dMat(prhs[0]);
    dMat  Y  = dMat(prhs[1]);
    dMat  W  = dMat(prhs[2]);
    dMat  V  = dMat(prhs[3]);
    dMat  weight = dMat(prhs[4]);
    
    plhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
    dMat L = dMat(plhs[0]);
    
    plhs[1] = mxCreateDoubleMatrix(W.M,W.N,mxREAL);
    dMat dW = dMat(plhs[1]);
    
    plhs[2] = mxCreateDoubleMatrix(V.M,V.N,mxREAL);
    dMat dV = dMat(plhs[2]);
    
    int NY = W.M;
    int NX = X.M;
    
    int nfeat = V.M;
    
    dMat P (W.M,1,0.0);
    dMat H (nfeat,1,0.0);
    dMat Hp(nfeat,1,0.0);
    dMat drez(W.M,1,0.0);
    dMat dH(nfeat,1,0.0);
    //dMat dV2(V.M,V.N,mxREAL);
    double expHy;
    for(n=0; n<X.N; n++){
        for(y=0; y<nfeat; y++){
            H(y)  =0;
            Hp(y) = 0;
            for(i=0; i<NX; i++)
                H(y) += V(y,i)*X(i,n);
            expHy = exp(-H(y));
            H( y)=1.0/(1+expHy) - .5; // sigmoid //-.5 for balance
            Hp(y)=expHy/(1+expHy)/(1+expHy);
        }

        for(y=0; y<NY; y++){
            P(y)=0;
            for(i=0; i<nfeat; i++)
                P(y) += W(y,i)*H(i);
            P(y)=exp(P(y));
        }
        P.norm_col(0);
                
        int myy = (int)Y(n)-1;
        L(0) += weight(n)*log(P(myy));

        // profiling shows a mere factor of two overhead for autodiff...
        for(i=0; i<H.M; i++)
            dW(myy,i) += weight(n)*H(i);
        for(y=0; y<W.M; y++)
            for(i=0; i<H.M; i++)
                dW(y,i) -= weight(n)*P(y)*H(i);

        for(y=0; y<W.M; y++)
            drez(y) = -P(y);
        drez(myy) += 1;
        for(i=0; i<V.M; i++){
            dH(i) = 0;
            for(y=0; y<W.M; y++)
                dH(i) += drez(y)*W(y,i);
        }       
        for(i=0; i<V.M; i++)
            for(j=0; j<V.N; j++)
                dV(i,j) += weight(n)*dH(i)*Hp(i)*X(j,n);
        
        // old method-- way slower
        /*
        // first, do yhat
        for(i=0; i<V.M; i++)
            for(j=0; j<V.N; j++)
                dV(i,j) += weight(n)*W(myy,i)*Hp(i)*X(j,n);
        // now do other elements
        for(y=0; y<W.M; y++)
            for(i=0; i<V.M; i++)
                for(j=0; j<V.N; j++)
                    dV(i,j) -= weight(n)*P(y)*W(y,i)*Hp(i)*X(j,n);
        */
    }
    return;
    
}

// this was one of those really freaky times in my life where IT WORKED
// THE FIRST TIME (2008.12.02, may I never forget thee.)
// actually upon further reflection, I needed to make the change
// V.M -> V.N on lines 90 and 95.  alas.
/* TESTO:
e=1e-6;
X = rand(10,100);
Y = ceil(5*rand(1,100));
weight = rand(1,100);
V = rand(10,10);
W = rand(5,10);
[L dW dV] = eval_fast_2layer2(X,Y,W,V,weight);
dW(1,1)
W2=W;W2(1,1)=W(1,1)+e; (eval_fast_2layer2(X,Y,W2,V,weight)-L)/e
dW(2,3)
W2=W;W2(2,3)=W(2,3)+e; (eval_fast_2layer2(X,Y,W2,V,weight)-L)/e
dV(1,1)
V2=V;V2(1,1)=V(1,1)+e; (eval_fast_2layer2(X,Y,W,V2,weight)-L)/e
dV(3,3)
V2=V;V2(3,3)=V(3,3)+e; (eval_fast_2layer2(X,Y,W,V2,weight)-L)/e
 */