/* Simple test program for the RAD package. */

/* Compute f(n,x) = the product of x[0], x[1], ..., x[n-1] */
/* and the gradient of f for two x values, and print the result. */

//#include "rad.hpp"
#include <stdio.h>
#include "mex.h"   //--This one is required
//#define nanoMatDebug 1 // 1 for safety/error checking; 0 for faster and unsafe(er)
#include "nanoMatT.cpp"
#include "math.h"
#define mmin(X,Y) ((X) < (Y) ? (X) : (Y))
#define mmax(X,Y) ((X) > (Y) ? (X) : (Y))

/*
typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<ADvar> adMat;
typedef ADvar var;
*/


typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<double> adMat;
typedef double var;

double my_tanh(double x,dMat *A){ // tabled tanh for speed...
    double where = (x+10)*100;
    where = mmax(mmin(where,2000),0);
    
    int floorwhere = (int)where;
    int ceilwhere  = mmin(2000,floorwhere+1);
    
    double frac_lo = ceilwhere-where;
    double frac_hi = 1-frac_lo;
    return (*A)(ceilwhere)*frac_hi + (*A)(floorwhere)*frac_lo;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){
    int i, j, k, l, m, n, y;
    
    dMat A(2001,1);
    for(i=0; i<2001; i++)
        A(i) = tanh(-10 + .01*i);
    
    dMat  X  = dMat(prhs[0]);
    dMat  Y  = dMat(prhs[1]);
    dMat  W  = dMat(prhs[2]);
    dMat  V  = dMat(prhs[3]);
    dMat  weight = dMat(prhs[4]);
    double lambda = dMat(prhs[5])(0);
    
    plhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
    dMat L = dMat(plhs[0]);
    
    plhs[1] = mxCreateDoubleMatrix(W.M,W.N,mxREAL);
    dMat dW = dMat(plhs[1]);
    
    plhs[2] = mxCreateDoubleMatrix(V.M,V.N,mxREAL);
    dMat dV = dMat(plhs[2]);
    
    int NY = W.M;
    int NX = X.M;
    
    int nfeat = V.M;
    
    dMat P (W.M,1,0.0);
    dMat H (nfeat,1,0.0);
    dMat Hp(nfeat,1,0.0);
    dMat drez(W.M,1,0.0);
    dMat dH(nfeat,1,0.0);
    double expHy;
    //double lambda = .05;
    for(int iter=0; iter<1; iter++){
        L(0) = 0;
        for(n=0; n<X.N; n++){
            double weight_n = weight(n);
            /*
            for(y=0; y<nfeat; y++){
                H(y)  =0;
                Hp(y) = 0;
                for(i=0; i<NX; i++)
                    H(y) += V(y,i)*X(i,n);
                H(y)  = tanh(H(y));
                Hp(y) = 1-H(y)*H(y);
            }
            H.printMe();
            Hp.printMe();
            */
            
            for(y=0; y<nfeat; y++)
                H(y)  =0;
            //for(i=0; i<NX; i++){
            //     double val = X(i,n);
            //     for(y=0; y<nfeat; y++)
            //         H(y) += V(y,i)*val;
            //}
            for(i=0; i<NX; i++){
                double *Vdata = V.data + V.M*i;
                double val = X(i,n);
                for(y=0; y<nfeat; y++)
                    H(y) += *(Vdata + y)*val;
            }
            for(y=0; y<nfeat; y++){
                Hp(y) = 0;
                //H(y)  = tanh(H(y));
                H(y)  = my_tanh(H(y),&A); // tabled tanh is faster!
                Hp(y) = 1-H(y)*H(y);
            }
            
            /*
            for(y=0; y<NY; y++){
                P(y)=0;
                for(i=0; i<nfeat; i++)
                    P(y) += W(y,i)*H(i);
                P(y)=exp(P(y));
            }
            P.norm_col(0);
            */
            
            for(y=0; y<NY; y++)
                P(y)=0;
            //for(i=0; i<nfeat; i++)
            //    for(y=0; y<NY; y++)
            //        P(y) += W(y,i)*H(i);
            for(i=0; i<nfeat; i++){
                double *Wdata = W.data + W.M*i;
                for(y=0; y<NY; y++)
                    P(y) += *(Wdata+y)*H(i);
            }
            for(y=0; y<NY; y++)
                P(y)=exp(P(y));
            P.norm_col(0);
            
            int myy = (int)Y(n)-1;
            L(0) += weight_n*log(P(myy));
            
            dW *= 0;
            for(i=0; i<H.M; i++)
                dW(myy,i) += weight_n*H(i);
            for(i=0; i<H.M; i++){
                double Hi = H(i);
                double* dWdata = dW.data + i*dW.M;
                for(y=0; y<W.M; y++)
                    *(dWdata+y) -= weight_n*P(y)*Hi;
                    //dW(y,i) -= weight(n)*P(y)*Hi;
            }
                
            for(y=0; y<W.M; y++)
                drez(y) = -P(y);
            drez(myy) += 1;
            for(i=0; i<V.M; i++){
                dH(i) = 0;
                for(y=0; y<W.M; y++)
                    dH(i) += drez(y)*W(y+i*W.M);
                    //dH(i) += drez(y)*W(y,i);
            }
            dV *= 0;
            for(j=0; j<V.N; j++){
                double *dVdata = dV.data+j*dV.M;
                double Xjn = X(j,n);
                for(i=0; i<V.M; i++)
                    *(dVdata+i) += weight_n*dH(i)*Hp(i)*Xjn;
                    //dV(i,j) += weight_n*dH(i)*Hp(i)*X(j,n);
            }
            
            // only do every 10 iterations because this is actually
            // pretty expensive
            if( (n % 1) == 0){
                double reglambda = .0001;
//                 for(i=0; i<W.M; i++)
//                     for(j=0; j<W.N; j++){
//                         L(0)    -= reglambda * W(i,j)*W(i,j);
//                         dW(i,j) -= reglambda*2*W(i,j);
//                     }
                for(i=0; i<W.M*W.N; i++){
                    L(0)  -= reglambda * W(i)*W(i);
                    dW(i) -= reglambda*2*W(i);
                }
                for(i=0; i<V.M*V.N; i++){
                    L(0)  -= reglambda * V(i)*V(i);
                    dV(i) -= reglambda*2*V(i);
                }
                dW -= W*(reglambda*2);
                dV -= V*(reglambda*2);
            }
            
            W += dW*lambda;
            V += dV*(lambda*4);
            
            //W += dW*.001; //.001 max
            //V += dV*.001; //.001 max
        }
        //printf("iter: %d  L: %f \n", iter, L(0));
    }
    return;
    
}