/* Simple test program for the RAD package. */

/* Compute f(n,x) = the product of x[0], x[1], ..., x[n-1] */
/* and the gradient of f for two x values, and print the result. */

//#include "rad.hpp"
#include <stdio.h>
#include "mex.h"   //--This one is required
//#define nanoMatDebug 1 // 1 for safety/error checking; 0 for faster and unsafe(er)
#include "nanoMatT.cpp"
#include "math.h"
#define mmin(X,Y) ((X) < (Y) ? (X) : (Y))
#define mmax(X,Y) ((X) > (Y) ? (X) : (Y))

/*
typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<ADvar> adMat;
typedef ADvar var;
*/

// function nodes = treeeval_helper(var,cut,children,X,At,mu)
// N = size(X,2);
// nodes = zeros(1,N);
// for n=1:N
//     x = X(:,n)-mu;
//     node = 1;
//     while var(node)
//         i    = var(node);
//         %xi   = a{i}*x;
//         xi   = At(i,:)*x;
//         if xi > cut(node)
//             win = 2;
//         else
//             win = 1;
//         end
//         node = children(node,win);
//     end
//     nodes(n) = node;
// end


typedef nanoMat<double> dMat;
typedef nanoMat<int>    iMat;
typedef nanoMat<double> adMat;
typedef double var;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){
    int i, j, k, l, m, n, y;
    
    dMat  var     = dMat(prhs[0]);
    dMat  cut     = dMat(prhs[1]);
    dMat children = dMat(prhs[2]);
    dMat   X      = dMat(prhs[3]);
    dMat   A      = dMat(prhs[4]);
    dMat   mu     = dMat(prhs[5]);
    
    int N = X.N;
    int M = X.M;
    int win;
    double xi;
    
    plhs[0] = mxCreateDoubleMatrix(1,N,mxREAL);
    dMat nodes = dMat(plhs[0]);
    
    /*
    dMat x = dMat(X.M,1,0);
    for(n=0; n<N; n++){
        for(m=0; m<M; m++)
            x(m)=X(m+n*M)-mu(m);
        int node=1;
        while(var(node-1)!=0){
            i = var(node-1);
            double xi = 0;
            for(m=0; m<M; m++)
                xi += A(m+(i-1)*M)*x(m);
            if(xi > cut(node-1))
                win = 2;
            else
                win = 1;
            node = children(node-1,win-1);
        }
        nodes(n)=node;
    }*/
    
//     // more clever version, caches components of AX
//     dMat x  = dMat(X.M,1,0);
//     dMat ax = dMat(A.N,1,0);
//     for(n=0; n<N; n++){
//         for(l=0; l<A.N; l++)
//             ax(l)=0;
//         for(m=0; m<M; m++)
//             x(m)=X(m+n*M)-mu(m);
//         int node=1;
//         while(var(node-1)!=0){
//             i = var(node-1);
//             if( ax(i-1)==0){
//                 xi = 0;
//                 for(m=0; m<M; m++)
//                     xi += A(m+(i-1)*M)*x(m);
//                 ax(i-1) = xi;
//             } else
//                 xi = ax(i-1);
//             if(xi > cut(node-1))
//                 win = 2;
//             else
//                 win = 1;
//             node = children(node-1,win-1);
//         }
//         nodes(n)=node;
//     }

    // more clever version, caches components of AX
    dMat x  = dMat(X.M,1,0);
    dMat ax = dMat(A.N,1,0);
    for(n=0; n<N; n++){
        for(l=0; l<A.N; l++)
            ax(l)=0;
        double *Xdata = X.data + n*M;
        double *xdata = x.data;
        double *mudata = mu.data;
        for(m=0; m<M; m++)
            *(xdata++) = *(Xdata++)-*(mudata++);
            //x(m)=X(m+n*M)-mu(m);
        int node=1;
        while(var(node-1)!=0){
            i = var(node-1);
            if( ax(i-1)==0){
                xi = 0;
                double *Adata = A.data+(i-1)*M;
                double *xdata = x.data;
                for(m=0; m<M; m++)
                    xi += *(Adata++) * *(xdata++);
                    //xi += *(Adata+m) * x(m);
                    //xi += A(m+(i-1)*M)*x(m);
                ax(i-1) = xi;
            } else
                xi = ax(i-1);
            if(xi > cut(node-1))
                win = 2;
            else
                win = 1;
            node = children(node-1,win-1);
        }
        nodes(n)=node;
    }    
    
    return;
    
}