% the following tricks were applied to make this system fast
% none of these is particularly clever or difficult, but putting
% everything together into a big system is tough
% - compiled c++ code for all bottlenecks
% - stochastic training
% - separable filters
% - tabling of tanh function for speed
% - quantizing and dequantizing data for compression when saving
% - reducing data size by PCA
% - computing PCA efficiently by creating covariance matrix and using eigs
% - cleverly stuffing PCA basis into first layer of mlp
% - all the big data in manipulated strictly in-place
% - in-place permutation
% - optimized (profiled) pointer manipulation in C++ code
%   (reordering loops, saving intermediate pointers, etc., etc.)


% MOST IMPORTANT PARAMETERS:
% num_scales
% num_training_elements (>= 1e5)
% 

% MEDIUM IMPORTANT
% num_features
% patch_size

% LESS IMPORATNT
% num_pca_comps

function train(imsdir,labelsdir,num_labels,model)

%rand('state',5);

patch_size = 7; % size of patches to use;
buf_size = (patch_size-1)/2;
num_training_elements = 1e5;%5e5; % total number of points to use to train
num_layers = 3;% 5 really
%num_labels = 5;
num_pca_comps = 100;%100 really

Aims    = dir([imsdir '/*.jpg']);
Alabels = dir([labelsdir '/*.png']);
% Aims = Alabels;

numims = length(Aims)

%numims = 30;
%for i=1:numims
%    Aims(i,1).name=regexprep(Aims(i,1).name,'png','jpg');
%end
fprintf('detected %d images.\n', numims);

num_colors = size(imread([imsdir    '/' Aims(1).name]),3);

points_per_im_pca     = ceil(num_training_elements/numims)

model.buf_size      = buf_size;
model.patch_size    = patch_size;
model.num_pca_comps = num_pca_comps;
model.num_labels    = num_labels;
model.type = 'mlp';
%model.type = 'dt';
num_features = 30;

num_scales = 5;
model.num_scales = num_scales;

%bigint = double(intmax('uint32'));
%resolution for saving-- allows matlab to compress much better
model.bigint = 2^16;

if nargin==4, layer0=model.numlayers+1; else, layer0=1; end

for layer=layer0:num_layers
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % go through the data, computing the mean and covariance matrix
    %
    % use linear algebra fact:
    % mean_i[(x_i-mu)*(x_i-mu)'] = mean_i[x_i*x_i'] - mu*mu'

    % for now, use ALL the data for computing the mean and cov matrix
    D  = num_scales*patch_size^2*num_colors + num_scales*double(layer>1)*patch_size^2*num_labels; %length of training vectors
    mu = zeros(D,1);
    C   = zeros(D);
    totpoints = 0;
    fprintf('going through data to build PCA...\n');
    tic
    im_scaled = []; P_scaled = [];
    for i=1:numims
        printstatus(i/numims,50,i~=1);
        [myX myY im_scaled P_scaled] = get_data_scaled(points_per_im_pca,imsdir,labelsdir,Aims,Alabels,i,layer,D,model,im_scaled,P_scaled);
        totpoints = totpoints+points_per_im_pca;
        mu = mu + sum(myX,2);
        C  = C  + myX*myX';
    end
    fprintf('%f seconds. \n', toc);
    mu = mu/totpoints;
    C  = C /totpoints;
    C  = C - mu*mu';
    fprintf('computing eigenvalues for pca...\n');
    options.disp = 0; % shut up eigs
    [model.A{layer} tmpD] = eigs(C,min(D,num_pca_comps),'lm',options);
    model.At{layer} = model.A{layer}';
    model.mu{layer} = mu;
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % gather data! (PCA it while gathering to save space...)
    %

    %%% first, decide how many training elements to take per image
    trains_per_im = ones(numims,1)*ceil(num_training_elements/numims);
    % pick a few images to get one less training element
    % so that trains_per_im sums exactly to num_training_elements
    extras = sum(trains_per_im)-num_training_elements;
    losers = randperm(numims); losers = losers(1:extras);
    trains_per_im(losers) = trains_per_im(losers)-1;

    %%% now actually gather the data
    % to keep a low memory profile, we never load more than one image at a time
    clear X; % just in case, make room
    %X = zeros(min(D,num_pca_comps),num_training_elements);
    X = ones(1+min(D,num_pca_comps),num_training_elements);
    Y = zeros(1,num_training_elements);
    fprintf('going through the data and gathering %d training points...\n',num_training_elements);
    where = 1;
    tic
    for i=1:numims
        %where_im = 1;
        %Xim = zeros(D,trains_per_im(i));
        printstatus(i/numims,50,i~=1);
        [myX myY im_scaled P_scaled] = get_data_scaled(trains_per_im(i),imsdir,labelsdir,Aims,Alabels,i,layer,D,model,im_scaled,P_scaled);
        Y(  where:where+trains_per_im(i)-1)=myY;
        mu = model.mu{layer}; At = model.At{layer};
        X(1:end-1,where:where+trains_per_im(i)-1)=At*(myX-repmat(mu,1,trains_per_im(i)));
        where = where + trains_per_im(i);
    end
    fprintf('%f seconds \n', toc);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % train a classifier
    %

    if strcmp(model.type,'lr')
        fprintf('training a logistic regression model...\n');
        model.W{layer} = fitlogreg(X,Y);
        model.numlayers = layer;
    elseif strcmp(model.type,'mlp')
        fprintf('training a multi-layer perceptron...\n');
        V = .01*randn(num_features,min(D,num_pca_comps)+1);
        W = .01*randn(num_labels,num_features);
        %[model.W{layer} model.V{layer}] = fitmlp(X,Y,[],W,V);
        %[model.W{layer} model.V{layer}] = fitmlp_stoch(X,Y,[],W,V,.01,100);
        %[model.W{layer} model.V{layer}] = fitmlp_stoch(X,Y,[],W,V,.002,.0001);
        %[model.W{layer} model.V{layer}] = fitmlp_stoch(X,Y,[],W,V,.01,.0001);

       % [W V] = fitmlp_stoch(X,Y,[],W,V,.001  ,20);
       % [W V] = fitmlp_stoch(X,Y,[],W,V,.0001 ,20);
       % [W V] = fitmlp_stoch(X,Y,[],W,V,.00001,20);

        [W V] = fitmlp_stoch(X,Y,[],W,V,.01   ,5);
        [W V] = fitmlp_stoch(X,Y,[],W,V,.001  ,5);
        [W V] = fitmlp_stoch(X,Y,[],W,V,.0001 ,5);
        [W V] = fitmlp_stoch(X,Y,[],W,V,.00001,5);
        
        model.W{layer}=W;
        model.V{layer}=V;
    elseif strcmp(model.type,'dt')
        fprintf('training a decision tree... ');
        tic
        splitmin = 100;
        %splitmin = ceil(num_training_elements/1000)
        model.T{layer} = treefit(X',Y','method','classification','splitmin',splitmin);
        fprintf('%f  seconds.\n',toc);
    else
        error('unsupported model type.');
    end
    model.numlayers = layer;
    
    save model model
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % run the classifier on all the data, and save the results
    %
    fprintf('running the classifier on all the training data...\n');
    tic
    L = 0;
    totpoints = 0;
    for i=1:numims
        filename=sprintf('TemporaryData/P%05d',i);
        
        im = imread([imsdir    '/' Aims(i).name]);
        im = double(im)/255;
        [ly lx lz] = size(im);
        label = imread([labelsdir '/' Alabels(i).name]);
        
        printstatus(i/numims,50,i~=1);
        P  = classify_image_scaled(im,model);
          
        
        for y=1:ly
            for x=1:lx
                if label(y,x)
                    L = L + log(P(y,x,label(y,x)));
                    totpoints = totpoints + 1;
                end
            end
        end
%         big_buf_size = buf_size*2^(num_scales-1);
%         for y=big_buf_size+1:ly-big_buf_size
%             for x=big_buf_size+1:lx-big_buf_size
%                 if label(y,x)
%                     L = L + log(P(y,x,label(y,x)));
%                     totpoints = totpoints + 1;
%                 end
%             end
%         end

        
        figure(2), miximshow(P); figure(1), imshow(im); drawnow
        % to save space, save it as a uint32 (plenty of resolution
        % there...)
        Pint=uint32(model.bigint*P); save(filename,'Pint');
        im_last = im;
    end
    fprintf('%f seconds\n',toc);
    fprintf('L/N on all training data: %f \n', L/totpoints);
end

end